.card-body {
    padding: 0.75rem;
}
pre {
    margin: 0;
    font-family: Courier,Consolas,"Liberation Mono","Courier New",monospace;
}
code {
    font-family: Courier,Consolas,"Liberation Mono","Courier New",monospace;
}
